#ifndef UUID_FD38CC81_DBD3_4204_A11C_7B1EBD750D4E
#define UUID_FD38CC81_DBD3_4204_A11C_7B1EBD750D4E

#include <cstddef>
#include <limits>
#include <string>
#include <vector>
#include <mpi.h>

namespace ghostmp
{

// broadcast the data with many "count" up to maximum vlaue of size_t.
// To use enhanced MPI_Bcast routine in some MPI environment such as K computer,
//   * restrict internal message size to multiple of 4
//   * avoid the use of MPI derived datatype
template <typename T>
int bcast(T* buffer, std::size_t count, int root, MPI_Comm comm);

template <typename T>
int send(T* buffer, std::size_t count, int dest, int tag, MPI_Comm comm);

template <typename T>
int recv(T* buffer, std::size_t count, int source, int tag, MPI_Comm comm,
    int* recv_source = 0, int* recv_tag = 0);

template <typename T>
int isend(T* buffer, std::size_t count, int dest, int tag, MPI_Comm comm,
    MPI_Request* request);

template <typename T>
int irecv(T* buffer, std::size_t count, int source, int tag, MPI_Comm comm,
    MPI_Request* request);

int send_tag(int dest, int tag, MPI_Comm comm);

int recv_tag(int source, int tag, MPI_Comm comm, MPI_Status* status);

const char* convert_thread_support_level_to_str(int support_level);

int create_large_contiguous_datatype(std::size_t elem_size, std::size_t count,
    MPI_Datatype oldtype, MPI_Datatype* newtype);

// read file using MPI_File_read_all
// return 1 if could not open file
// return 2 if could not read file
// return 0 otherwise
template <typename T>
int read_all(const char* filename, std::vector<T>& output, MPI_Comm comm);

// write output using MPI_File_write_all
// return 1 if could not open file
// return 2 if could not write file
// return 0 otherwise
template <typename T>
int write_all(const char* filename, const std::vector<T>& output, MPI_Comm comm);

// read fasta file using MPI_File_read_all
// return 1 if could not open file
// return 2 if could not read file
// return 3 if data, result of read_all function (MPI_File_read_all),
//  contain no beginning of fasta entries.
// return 0 otherwise
int read_fasta(const char* filename, std::string& fasta, MPI_Comm comm);

} // namespace ghostmp

// ------------------------------ IMPLEMENTATION ------------------------------

namespace ghostmp
{

namespace detail
{

const int mn_for_fast_comm = 4; // magic number for fast communication
template <typename T> MPI_Datatype get_mpi_datatype();

#define MPI_UTIL_MAKE_GET_MPI_DATATYPE(MUMGMD_TYPE, MUMGMD_MPI_TYPE) \
template <> inline \
MPI_Datatype get_mpi_datatype<MUMGMD_TYPE>() \
{ \
  return MUMGMD_MPI_TYPE; \
}

MPI_UTIL_MAKE_GET_MPI_DATATYPE(void, MPI_BYTE)
MPI_UTIL_MAKE_GET_MPI_DATATYPE(char, MPI_CHAR)
MPI_UTIL_MAKE_GET_MPI_DATATYPE(short, MPI_SHORT)
MPI_UTIL_MAKE_GET_MPI_DATATYPE(int, MPI_INT)
MPI_UTIL_MAKE_GET_MPI_DATATYPE(long, MPI_LONG)
MPI_UTIL_MAKE_GET_MPI_DATATYPE(float, MPI_FLOAT)
MPI_UTIL_MAKE_GET_MPI_DATATYPE(double, MPI_DOUBLE)
MPI_UTIL_MAKE_GET_MPI_DATATYPE(unsigned char, MPI_UNSIGNED_CHAR)
MPI_UTIL_MAKE_GET_MPI_DATATYPE(unsigned short, MPI_UNSIGNED_SHORT)
MPI_UTIL_MAKE_GET_MPI_DATATYPE(unsigned, MPI_UNSIGNED)
MPI_UTIL_MAKE_GET_MPI_DATATYPE(unsigned long, MPI_UNSIGNED_LONG)
#undef MPI_UTIL_MAKE_GET_MPI_DATATYPE

template <typename T, typename U>
struct IsSameType { static const bool value = false; };

template <typename T>
struct IsSameType<T, T> { static const bool value = true; };

template <typename T>
struct IsMpiBasicType
{
  static const bool value = (
      IsSameType<T, void>::value ||
      IsSameType<T, char>::value ||
      IsSameType<T, short>::value ||
      IsSameType<T, int>::value ||
      IsSameType<T, long>::value ||
      IsSameType<T, float>::value ||
      IsSameType<T, double>::value ||
      IsSameType<T, unsigned char>::value ||
      IsSameType<T, unsigned short>::value ||
      IsSameType<T, unsigned int>::value ||
      IsSameType<T, unsigned long>::value);
};

template <bool Condition> struct enable_if;

template <> struct enable_if<true> { typedef void type; };

template <bool Condition> struct disable_if;

template <> struct disable_if<false> {typedef void type; };

} // namespace detail

template <typename T>
int bcast(T* buffer, std::size_t count, int root, MPI_Comm comm)
{
  // divide "buffer" into "body" and "tail" to make the counts of
  // messages multiple of 4.
  T* body = reinterpret_cast<T*>(buffer);
  int rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  T tail[detail::mn_for_fast_comm];
  std::size_t body_count = (count / detail::mn_for_fast_comm) * detail::mn_for_fast_comm;
  int tail_count = count - body_count;
  if (rank == root && tail_count != 0) {
    std::copy(body + body_count, body + count, tail);
  }

  std::size_t i = 0;
  const int block_size = (std::numeric_limits<int>::max() /
      detail::mn_for_fast_comm) * detail::mn_for_fast_comm;
  for (; i + block_size < body_count; i += block_size) {
    T* p = body + i;
    int rc = MPI_Bcast(p, block_size, detail::get_mpi_datatype<T>(), root, comm);
    if (rc != MPI_SUCCESS) {
      return rc;
    }
  }
  if (i < body_count) {
    T* p = body + i;
    int rc = MPI_Bcast(p, body_count - i, detail::get_mpi_datatype<T>(), root, comm);
    if (rc != MPI_SUCCESS) {
      return rc;
    }
  }

  if (tail_count != 0) {
    int rc = MPI_Bcast(tail, detail::mn_for_fast_comm, detail::get_mpi_datatype<T>(), root,
        comm);
    if (rc != MPI_SUCCESS) {
      return rc;
    }
    if (rank != root) {
      std::copy(tail, tail + tail_count, body + body_count);
    }
  }
  return MPI_SUCCESS;
}

template <typename T>
int send(T* buffer, std::size_t count, int dest, int tag, MPI_Comm comm)
{
  int rc;
  if (count == 0) {
    rc = MPI_Send(buffer, count, detail::get_mpi_datatype<T>(), dest, tag, comm);
  } else {
    MPI_Datatype newtype;
    rc = create_large_contiguous_datatype(sizeof(T), count,
        detail::get_mpi_datatype<T>(), &newtype);
    if (rc != MPI_SUCCESS) { return rc; }
    rc = MPI_Type_commit(&newtype);
    if (rc != MPI_SUCCESS) { return rc; }
    rc = MPI_Send(buffer, 1, newtype, dest, tag, comm);
    if (rc != MPI_SUCCESS) { return rc; }
    rc = MPI_Type_free(&newtype);
    if (rc != MPI_SUCCESS) { return rc; }
  }
  return rc;
}

template <typename T>
int recv(T* buffer, std::size_t count, int source, int tag, MPI_Comm comm,
    int* recv_source, int* recv_tag)
{
  int rc;
  MPI_Status status;
  if (count == 0) {
    rc = MPI_Recv(buffer, count, detail::get_mpi_datatype<T>(), source, tag, comm,
        &status);
    if (rc != MPI_SUCCESS) { return rc; }
  } else {
    MPI_Datatype newtype;
    rc = create_large_contiguous_datatype(sizeof(T), count,
        detail::get_mpi_datatype<T>(), &newtype);
    if (rc != MPI_SUCCESS) { return rc; }
    rc = MPI_Type_commit(&newtype);
    if (rc != MPI_SUCCESS) { return rc; }
    rc = MPI_Recv(buffer, 1, newtype, source, tag, comm, &status);
    if (rc != MPI_SUCCESS) { return rc; }
    rc = MPI_Type_free(&newtype);
    if (rc != MPI_SUCCESS) { return rc; }
  }
  if (recv_source != NULL) {
    *recv_source = status.MPI_SOURCE;
  }
  if (recv_tag != NULL) {
    *recv_tag = status.MPI_TAG;
  }
  return rc;
}

template <typename T>
int isend(T* buffer, std::size_t count, int dest, int tag, MPI_Comm comm,
    MPI_Request* request)
{
  int rc;
  if (count == 0) {
    rc = MPI_Isend(buffer, count, detail::get_mpi_datatype<T>(), dest, tag, comm,
        request);
  } else {
    MPI_Datatype newtype;
    rc = create_large_contiguous_datatype(sizeof(T), count,
        detail::get_mpi_datatype<T>(), &newtype);
    if (rc != MPI_SUCCESS) { return rc; }
    rc = MPI_Type_commit(&newtype);
    if (rc != MPI_SUCCESS) { return rc; }
    rc = MPI_Isend(buffer, 1, newtype, dest, tag, comm, request);
    if (rc != MPI_SUCCESS) { return rc; }
    rc = MPI_Type_free(&newtype);
    if (rc != MPI_SUCCESS) { return rc; }
  }
  return rc;
}

template <typename T>
int irecv(T* buffer, std::size_t count, int source, int tag, MPI_Comm comm,
    MPI_Request* request)
{
  int rc;
  MPI_Status status;
  if (count == 0) {
    rc = MPI_Recv(buffer, count, detail::get_mpi_datatype<T>(), source, tag, comm,
        &status);
    if (rc != MPI_SUCCESS) { return rc; }
  } else {
    MPI_Datatype newtype;
    rc = create_large_contiguous_datatype(sizeof(T), count,
        detail::get_mpi_datatype<T>(), &newtype);
    if (rc != MPI_SUCCESS) { return rc; }
    rc = MPI_Type_commit(&newtype);
    if (rc != MPI_SUCCESS) { return rc; }
    rc = MPI_Irecv(buffer, 1, newtype, source, tag, comm, request);
    if (rc != MPI_SUCCESS) { return rc; }
    rc = MPI_Type_free(&newtype);
    if (rc != MPI_SUCCESS) { return rc; }
  }
  return rc;
}

template <typename T>
int read_all(const char* filename, std::vector<T>& data, MPI_Comm comm)
{
  int rank;
  MPI_Comm_rank(comm, &rank);
  int nprocs;
  MPI_Comm_size(comm, &nprocs);
  int worker_size = nprocs - 1;

  int rc;
  MPI_File fh;
  rc = MPI_File_open(comm, const_cast<char*>(filename), MPI_MODE_RDONLY, MPI_INFO_NULL, &fh);
  if (rc != MPI_SUCCESS) { return 1; }
  MPI_Offset file_size;
  rc = MPI_File_get_size(fh, &file_size);
  if (rc != MPI_SUCCESS) { return 2; }
  int residual =  file_size % worker_size;
  MPI_Offset offset = file_size / worker_size * (rank - 1) + (rank - 1 >= residual ? residual : rank - 1);
  if (rank == 0) {
    offset = 0;
  }
  rc = MPI_File_seek(fh, offset, MPI_SEEK_SET);
  if (rc != MPI_SUCCESS) { return 2; }
  int read_size = file_size / worker_size + (rank - 1 < residual ? 1 : 0);
  if (rank == 0) {
    read_size = 0;
  }
  data.resize(read_size);
  rc = MPI_File_read_all(fh, &data[0], data.size(), MPI_CHAR, MPI_STATUS_IGNORE);
  if (rc != MPI_SUCCESS) { return 2; }
  rc = MPI_File_close(&fh);
  if (rc != MPI_SUCCESS) { return 2; }
  return 0;
}

template <typename T>
int write_all(const char* filename, const std::vector<T>& data, MPI_Comm comm)
{
  int rank;
  MPI_Comm_rank(comm, &rank);
  int nprocs;
  MPI_Comm_size(comm, &nprocs);

  int rc;
  MPI_File fh;
  rc = MPI_File_open(comm, filename, MPI_MODE_WRONLY | MPI_MODE_CREATE | MPI_MODE_EXCL, MPI_INFO_NULL, &fh);
  if (rc != MPI_SUCCESS) { return 1; }
  std::size_t data_size = data.size();
  std::size_t size_prefix_sum;
  rc = MPI_Scan(&data_size, &size_prefix_sum, 1, MPI_UNSIGNED_LONG, MPI_SUM, comm);
  if (rc != MPI_SUCCESS) { return 2; }
  MPI_Offset offset = size_prefix_sum - data.size();
  rc = MPI_File_seek(fh, offset, MPI_SEEK_SET);
  if (rc != MPI_SUCCESS) { return 2; }
  rc = MPI_File_write_all(fh, &data[0], data.size(), MPI_CHAR, MPI_STATUS_IGNORE);
  if (rc != MPI_SUCCESS) { return 2; }
  rc = MPI_File_close(&fh);
  if (rc != MPI_SUCCESS) { return 2; }

  return 0;
}

} // namespace ghostmp

#endif

