#include <cstdlib>
#include <cstring>
#include <exception>
#include <iostream>

#include <mpi.h>

#include "ghostmp_common_parameters.h"
#include "common_parameters.h"
#include "aligner.h"
#include "aligner_parameters.h"
#include "ghostmp_mpi_util.h"

void PrintUsage() {
  std::cerr <<
    "ghostmp_search - parallel homology search tool. version "
    + ghostmp::kVersion + ".\n"
    "(based on GHOSTX-" + CommonParameters::kVersion +
    ", serial homology search tool.)\n" << std::endl;
  std::cerr <<
    "  ghostmp_search [-i queries] [-o output] [-d databes] [-v maxNumAliSub]\n"
    "                 [-b maxNumAliQue] [-M scoreMatrix] [-G openGap] [-E extendGap]\n"
    "                 [-l CandidatesSize] [-s lowerCutoff] [-T UpperCutoff]\n"
    "                 [-S searchLength] [-q queryType] [-t databaseType]\n"
    "                 [-a numThreads] [-L maxNumHits] [-w maxAliLen]\n"
    "\n"
    "  Options:\n"
    "  (Required)\n"
    "    -i STR    Input query name (must be formatted)\n"
    "    -o STR    Output file\n"
    "    -d STR    database name (must be formatted)\n"
    "\n"
    "  (Optional)\n"
    "    -v INT    Maximum number of alignments for each subject [1]\n"
    "    -b INT    Maximum number of the output for a query [10]\n"
    "\n"
    "    -M STR    Score matrix file[BLOSUM62]\n"
    "    -G INT    Open gap penalty [11]\n"
    "    -E INT    Extend gap penalty [1]\n"
    "\n"
    "    -l INT    Maximun size of the candidates (Bytes) [134217728 (=128MB)]\n"
    "    -s INT    Lower limit cutoff score for seed search [4]\n"
    "    -T INT    Upper limit cutoff score for seed search [30]\n"
    "    -S INT    Maximum length of alignments in seed search [10]\n"
    "    -q STR    Query sequence type, p (protein) or d (dna) [p]\n"
    "    -t STR    Database sequence type, p (protein) or d (dna) [p]\n"
    "    -F STR    Filter query sequence, T (enable) or F (disable) [T] \n" 
    "    -a INT    The number of threads [1]\n"
    "    -L INT    Maximum number of hits [67108864]"
    << std::endl;
}

int main(int argc, char* argv[]) {
  int required = MPI_THREAD_FUNNELED;
  int provided;
  MPI_Init_thread(&argc, &argv, required, &provided);
  int rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  int nprocs;
  MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
  if (provided < required) {
    const char* provided_str = ghostmp::convert_thread_support_level_to_str(provided);
    if (rank == 0) {
      std::cerr << "Error: provided level of thread support of this MPI environment is too low ("
        << provided_str << ')' << std::endl;
    }
    MPI_Finalize();
    exit(EXIT_FAILURE);
  }

  if (argc < 2 ||
      strcmp(argv[1], "-h") == 0 ||
      strcmp(argv[1], "--help") == 0) {
    if (rank == 0) {
      PrintUsage();
    }
    MPI_Finalize();
    exit(EXIT_FAILURE);
  }
  try {
    AlignerParameters parameters;
    parameters.SetParameters(argc, argv);
    if (nprocs == 1 ) {
      std::cerr << "not implemented yet" << std::endl;
    } else {
      Aligner aligner;
      aligner.Run(parameters);
    }
  } catch (std::exception &e) {
    std::cerr << e.what();
    MPI_Abort(MPI_COMM_WORLD, EXIT_FAILURE);
  }

  MPI_Finalize();
  return EXIT_SUCCESS;
}
