/*
 * searcher.cpp
 *
 *  Created on: 2010/09/16
 *      Author: shu
 */

#include "searcher.h"
#include "encoder.h"
#include "aligner_parameters.h"
#include "gapped_extender.h"
#include <iostream>
#include <stdint.h>
#include <limits.h>
#include <stdlib.h>
#include <math.h>
#include <string>
#include <vector>
#include <stdexcept>
#include <algorithm>
#include <assert.h>

using namespace std;

class AlignmentQueryComp {
public:
  bool operator()(const Alignment &a1, const Alignment &a2) const {
    return (a1.hit.queries_offset < a2.hit.queries_offset);
  }
};

Searcher::Searcher(const AlignerParameters &parameters) :
  max_regular_letter_(parameters.GetAlignedType()->GetRegularLetters().length()
      - 1), search_length_(parameters.GetSearchLength()), threshold_(
      parameters.GetThreshold()), cutoff_(parameters.GetSearchCutoff()), max_number_alignment_(
          parameters.GetMaxNumberHits()), score_matrix_(
          parameters.GetScoreMatrix()), matrix_(score_matrix_.GetMatrix()),
      query_index_(NULL), db_index_(NULL),
      query_length_(0), db_length_(0), query_seq_(NULL),
      db_seq_(NULL), ungapped_extend_nominal_cutoff_(0), gapped_extend_nominal_trigger_(0),
      query_start_list_(0), query_end_list_(0), db_start_list_(0), db_end_list_(0)
{
  for (uint32_t i = 0; i < search_length_; ++i) {
    score_cutoff_list_.push_back(0);
  }
}

Searcher::~Searcher() {
}

void Searcher::Set(Query &query, Db &db) {
  query_seq_ = query.GetSequence();
  query_length_ = query.GetSequenceLength();
  query_index_ = query.GetIndex();
  ungapped_extend_nominal_cutoff_ = -query.GetUngappedExtendNominalCutoff();
  gapped_extend_nominal_trigger_ = query.GetGappedExtendNominalTrigger();

  db_index_ = db.GetIndex();
  db_length_ = db.GetConcatenatedSequenceLength();
  db_seq_ = db.GetConcatenatedSequence();

  InitNodeStack();
}

void Searcher::Search(vector<Hit> &hits) {
  hits.clear();
  if (query_seq_ == NULL || query_index_ == NULL) {
    throw std::domain_error("error:not set query.");
  }
  while (!node_stack_.empty()) {
    Node node = node_stack_.top();
#if 0
    uint32_t stack_size = node_stack_.size();
    uint32_t query_pos = query_index_->Locate(node.query_start) + node.offset;
    uint32_t db_pos = db_index_->Locate(node.db_start) + node.offset;
#endif
    if (node.score > threshold_ && kMinSearchLength < node.offset) {
      node_stack_.pop();
      uint32_t query_count = node.query_end - node.query_start + 1;
      if (query_count >= max_number_alignment_) {
        query_count = max_number_alignment_ - 1;
        Node new_node = node;
        uint32_t query_start_s = node.query_start;
        uint32_t queries_start_e = node.query_start + query_count - 1;
        while (queries_start_e < node.query_end) {
          new_node.query_start = query_start_s;
          new_node.query_end = queries_start_e;
          node_stack_.push(new_node);
          query_start_s = queries_start_e + 1;
          queries_start_e = query_start_s + query_count - 1;
        }
        new_node.query_start = query_start_s;
        new_node.query_end = node.query_end;
        node_stack_.push(new_node);
        node = node_stack_.top();
        query_count = node.query_end - node.query_start + 1;
      }

      Hit h;
      h.score = 0;
      while ((hits.size() + query_count < max_number_alignment_)
          && node.db_start <= node.db_end) {
        h.start.db_offset = db_index_->Locate(node.db_start);
        h.hit.db_offset = h.start.db_offset + node.offset;
        h.end.db_offset = h.hit.db_offset;
        for (uint32_t i = node.query_start; i <= node.query_end; ++i) {
          h.start.queries_offset = query_index_->Locate(i);
          h.hit.queries_offset = h.start.queries_offset + node.offset;
          h.end.queries_offset = h.hit.queries_offset;
	  //hits.push_back(h);
	  
          if (node.score >= gapped_extend_nominal_trigger_) {
            hits.push_back(h);
          } else if (Filter(h.hit.queries_offset, h.hit.db_offset)) {
            hits.push_back(h);
          }
	  
	 
        }
        ++node.db_start;
      }
      if (node.db_start <= node.db_end) {
        node_stack_.push(node);
        return;
      }
    } else {
      node_stack_.pop();
      uint32_t offset = node.offset + 1;
      if (offset < search_length_) {
        SetNextNodes(node);
      }
    }
  }
  return;
}

void Searcher::InitNodeStack() {
  while (!node_stack_.empty()) {
    node_stack_.pop();
  }
  uint8_t query_start_character = 0;
  uint8_t query_end_character = max_regular_letter_;
  uint8_t db_start_character = 0;
  uint8_t db_end_character = max_regular_letter_;
  int *matrix = &matrix_[0];
  uint32_t number_letters = score_matrix_.GetNumberLetters();

  query_start_list_.resize(query_index_->GetMaxAlphabet() + 1);
  query_end_list_.resize(query_index_->GetMaxAlphabet() + 1);
  db_start_list_.resize(db_index_->GetMaxAlphabet() + 1);
  db_end_list_.resize(db_index_->GetMaxAlphabet() + 1);
  
  query_index_->Search(0, 0, query_length_ - 1, &query_start_character, &query_end_character, &query_start_list_[0], &query_end_list_[0]);
  if (query_start_list_[query_start_character] > query_end_list_[query_end_character]) {
    return;
  }

  db_index_->Search(0, 0, db_length_ - 1, &db_start_character, &db_end_character, &db_start_list_[0], &db_end_list_[0]);
  if (db_start_list_[db_start_character] > db_end_list_[db_end_character]) {
    return;
  }

  for (uint32_t i = query_start_character; i <= query_end_character; ++i) {
    if (query_start_list_[i] <= query_end_list_[i]) {
      Node next_node;
      next_node.query_start = query_start_list_[i];
      next_node.query_end = query_end_list_[i];
      int *score_matrix_row = matrix + number_letters * i;
      next_node.max_score = score_matrix_row[i];
      int cutoff = max(next_node.max_score - cutoff_, score_cutoff_list_[0]);
      for (uint32_t j = db_start_character; j <= db_end_character; ++j) {
        next_node.db_start = db_start_list_[j];
        next_node.db_end = db_end_list_[j];
        int s = score_matrix_row[j];
        if ((next_node.db_start <= next_node.db_end) && (s > cutoff)) {
#if 0
          uint32_t query_pos = query_index_->Locate(next_node.query_start);
          uint32_t db_pos = db_end_ - db_index_->Locate(next_node.db_start);
#endif
          next_node.score = s;
          next_node.offset = 0;
          node_stack_.push(next_node);
        }
      }
    }
  }
}

void Searcher::SetNextNodes(const Node &node) {
  uint8_t query_start_character = 0;
  uint8_t query_end_character = max_regular_letter_;
  uint8_t db_start_character = 0;
  uint8_t db_end_character = max_regular_letter_;
  uint32_t offset = node.offset + 1;
  int *matrix = &matrix_[0];
  uint32_t number_letters = score_matrix_.GetNumberLetters();

  query_start_list_.resize(query_index_->GetMaxAlphabet() + 1);
  query_end_list_.resize(query_index_->GetMaxAlphabet() + 1);
  db_start_list_.resize(db_index_->GetMaxAlphabet() + 1);
  db_end_list_.resize(db_index_->GetMaxAlphabet() + 1);


  query_index_->Search(offset, node.query_start, node.query_end, &query_start_character, &query_end_character, &query_start_list_[0], &query_end_list_[0]);
  if (query_start_list_[query_start_character] > query_end_list_[query_end_character]) {
    return;
  }

  db_index_->Search(offset, node.db_start, node.db_end, &db_start_character, &db_end_character, &db_start_list_[0], &db_end_list_[0]);
  if (db_start_list_[db_start_character] > db_end_list_[db_end_character]) {
    return;
  }

  for (uint32_t i = query_start_character; i <= query_end_character; ++i) {
    if (query_start_list_[i] <= query_end_list_[i]) {
      Node next_node;
      next_node.query_start = query_start_list_[i];
      next_node.query_end = query_end_list_[i];
      next_node.offset = offset;
      int *score_matrix_row = matrix + number_letters * i;
      next_node.max_score = node.max_score + score_matrix_row[i];
      int cutoff = max(next_node.max_score - cutoff_, score_cutoff_list_[offset]);
      for (uint32_t j = db_start_character; j <= db_end_character; ++j) {
        next_node.db_start = db_start_list_[j];
        next_node.db_end = db_end_list_[j];
        next_node.score = node.score + score_matrix_row[j];
        if (next_node.db_start <= next_node.db_end && next_node.score > cutoff) {
          node_stack_.push(next_node);
        }
      }
    }
  }
}

bool Searcher::Filter(uint32_t queries_offset, uint32_t db_offset) {
  uint32_t q_offset = queries_offset;
  uint32_t d_offset = db_offset;
  int score = 0;
  int s = 0;
  int cutoff = ungapped_extend_nominal_cutoff_;
  uint32_t number_letters = score_matrix_.GetNumberLetters();
  int *score_matrix = &(matrix_[0]);

  //cout << "ungapped extend" << endl;
  //cout << queries_offset << " " << db_offset  << endl;

  do {
    if (db_seq_[d_offset] == AlignerParameters::kSequenceEnd
        || query_seq_[q_offset] == AlignerParameters::kSequenceEnd) {
      break;
    }
    //cout << q_offset << " " << d_offset << " " <<(int)query_seq_[q_offset] << " " << (int)db_seq_[d_offset] << " "  <<  score <<endl;
    s += score_matrix[db_seq_[d_offset] * number_letters + query_seq_[q_offset]];
    --d_offset;
    --q_offset;
    if (s > 0) {
      do {
        score += s;
        if (score >= gapped_extend_nominal_trigger_) {
	  return true;
        }
        if (db_seq_[d_offset] == AlignerParameters::kSequenceEnd
            || query_seq_[q_offset] == AlignerParameters::kSequenceEnd) {
          break;
        }
	//cout << q_offset << " " << d_offset << " " <<(int)query_seq_[q_offset] << " " << (int)db_seq_[d_offset] << " "  <<  score <<endl;
        s = score_matrix[db_seq_[d_offset] * number_letters + query_seq_[q_offset]];
        --d_offset;
        --q_offset;
      } while (s > 0);
    }
  } while (s > cutoff);

  q_offset = queries_offset + 1;
  d_offset = db_offset + 1;
  s = 0;
  if (-score > ungapped_extend_nominal_cutoff_) {
    cutoff = -score;
  }
  //cout << "-"<< endl;

  do {
    if (db_seq_[d_offset] == AlignerParameters::kSequenceEnd
        || query_seq_[q_offset] == AlignerParameters::kSequenceEnd) {
      break;
    }
    //cout << q_offset << " " << d_offset << " " <<(int)query_seq_[q_offset] << " " << (int)db_seq_[d_offset] << " "  <<  score <<endl;
    s += score_matrix[db_seq_[d_offset] * number_letters + query_seq_[q_offset]];
    ++d_offset;
    ++q_offset;
    if (s > 0) {
      do {
        score += s;
        if (score >= gapped_extend_nominal_trigger_) {
          return true;
        }
        if (db_seq_[d_offset] == AlignerParameters::kSequenceEnd
            || query_seq_[q_offset] == AlignerParameters::kSequenceEnd) {
          break;
        }
	//cout << q_offset << " " << d_offset << " " <<(int)query_seq_[q_offset] << " " << (int)db_seq_[d_offset] << " "  <<  score <<endl;
        s = score_matrix[db_seq_[d_offset] * number_letters
            + query_seq_[q_offset]];
        ++d_offset;
        ++q_offset;
      } while (s > 0);
      if ((cutoff = -score) < ungapped_extend_nominal_cutoff_) {
        cutoff = ungapped_extend_nominal_cutoff_;
      }
    }
  } while (s > cutoff);

  return false;
}


