/*
 * suffix_array.cpp
 *
 *  Created on: 2010/08/02
 *      Author: shu
 */

#include <new>
#include <fstream>
#include <iostream>
#include <vector>
#include <assert.h>
#include <cstdlib>
#include "sais.h"
#include "suffix_array.h"

using namespace std;

SuffixArray::SuffixArray()
:max_alphabet_(0), text_(0), suffix_array_(0)
{}

SuffixArray::SuffixArray(const unsigned char *text, unsigned int length)
:max_alphabet_(0), text_(0), suffix_array_(0)
{
  Rebuild(text, length);
}

SuffixArray::SuffixArray(ifstream &is) {
  Load(is);
}

void SuffixArray::Rebuild(const unsigned char *text, unsigned int length) {
  vector<int> sa(length);
  for (unsigned int i = 0; i < length; ++i) {
    if (max_alphabet_ < text[i]) {
      max_alphabet_ = text[i];
    }
  }

  text_.insert(text_.begin(), text, text + length);


  sais(&text_[0], &sa[0], text_.size());
  suffix_array_.insert(suffix_array_.begin(), sa.begin(), sa.end());
  assert(text_.size() == suffix_array_.size());
}

void SuffixArray::Search(unsigned char character, unsigned int offset,
    unsigned int *start, unsigned int *end) {
  if (character > max_alphabet_) {
    *start = 1;
    *end = 0;
    return;
  }

  if (suffix_array_[*start] + offset >= text_.size()) {
    ++(*start);
  }

  unsigned int s = *start;
  unsigned int e = *end;
  unsigned int m;
  unsigned char *t = &text_[0] + offset;

  if (s > e) {
    *start = 1;
    *end = 0;
    return;
  } else if (s == e) {
    if (t[suffix_array_[s]] == character) {
      return;
    } else {
      *start = 1;
      *end = 0;
      return;
    }
  }

  if (t[suffix_array_[s]] != character) {
    while (s < e - 1) {
      m = (s + e) / 2;
      if (t[suffix_array_[m]] < character) {
        s = m;
      } else {
        e = m;
      }
    }
    if (t[suffix_array_[e]] != character) {
      *start = 1;
      *end = 0;
      return;
    }
    *start = s = e;
    e = *end;
  }

  if (t[suffix_array_[e]] != character) {
    while (s < e - 1) {
      m = (s + e) / 2;
      if (t[suffix_array_[m]] > character) {
        e = m;
      } else {
        s = m;
      }
    }
    if (t[suffix_array_[s]] != character) {
      *start = 1;
      *end = 0;
      return;
    }
    *end = s;
  }
  return;
}

void SuffixArray::Search(unsigned int offset, unsigned int start,
    unsigned int end, unsigned char *start_character,
    unsigned char *end_character, unsigned int *start_list, unsigned int *end_list) {
  unsigned char start_c = *start_character;
  unsigned char end_c = *end_character;

  if (suffix_array_[start] + offset >= text_.size()) {
    ++start;
  }
  if (start > end) {
    *end_character = start_c;
    start_list[start_c] = 1;
    end_list[start_c] = 0;
    return;
  } else if (start == end) {
    unsigned char c = text_[offset +suffix_array_[start]];
    if (start_c <= c && c <= end_c) {
      start_list[c] = start;
      end_list[c] = start;
      *start_character = c;
      *end_character = c;
    } else {
      *end_character = start_c;
      start_list[start_c] = 1;
      end_list[start_c] = 0;
    }
    return;
  }

  unsigned int *searched_list;
  const unsigned int sl_array_size = 258;
  unsigned int sl_array[sl_array_size];
  unsigned char *t = &text_[offset];
  unsigned char searched_end_c = end_c;
  
  if (sl_array_size < max_alphabet_ + (unsigned int)2) {
    searched_list = new unsigned int[max_alphabet_ + 2];
  } else {
    searched_list = sl_array;
  }

  if (t[suffix_array_[start]] > start_c) {
    start_c = t[suffix_array_[start]];
    *start_character = start_c;
  }

  if (t[suffix_array_[end]] <= end_c) {
    end_c = t[suffix_array_[end]];
    searched_list[end_c + 1] = end + 1;
    *end_character = end_c;
    searched_end_c = end_c;
  } else {
    ++searched_end_c;
  }

  searched_list[start_c] = start;

  for (unsigned char i = start_c + 1; i <= searched_end_c; ++i) {
    unsigned int s = searched_list[i - 1];
    if (t[suffix_array_[s]] >= i) {
      searched_list[i] = s;
    } else {
      unsigned int e = end;
      unsigned int m;
      while (s < e - 1) {
        m = (s + e) / 2;
        if (t[suffix_array_[m]] < i) {
          s = m;
        } else {
          e = m;
        }
      }
      searched_list[i] = e;
    }
  }

  for (unsigned char i = start_c; i <= end_c; ++i) {
    start_list[i] = searched_list[i];
    end_list[i] = searched_list[i + 1] - 1;
  }
  if (searched_list != sl_array) {
    delete [] searched_list;
  }
}

unsigned int SuffixArray::Locate(unsigned int i) {
  return suffix_array_[i];
}

void SuffixArray::Load(ifstream &is) {
  unsigned int length = 0;
  is.read((char*)(&max_alphabet_), sizeof(max_alphabet_));
  is.read((char*)(&length), sizeof(length));
  text_.resize(length);
  suffix_array_.resize(length);
  is.read((char*)(&text_[0]), sizeof(text_[0])*length);
  is.read((char*)(&suffix_array_[0]), sizeof(suffix_array_[0])*length);
}

void SuffixArray::Save(ofstream &os) {
  unsigned int length = text_.size();
  os.write((const char*)(&max_alphabet_), sizeof(max_alphabet_));
  os.write((const char*)(&length), sizeof(length));
  os.write((const char*)(&text_[0]), sizeof(text_[0])*length);
  os.write((const char*)(&suffix_array_[0]), sizeof(suffix_array_[0])*length);
}

